//
// Copyright (c) 2008, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   12 Jun 08  Brian Frank  Creation
//
package fan.fwt;

import java.util.ArrayList;
import fan.sys.*;
import fan.sys.List;
import org.eclipse.swt.*;
import org.eclipse.swt.custom.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Widget;
import org.eclipse.swt.events.*;

public class RichTextPeer
  extends WidgetPeer
  implements LineStyleListener, ModifyListener
{

  public Widget create(Widget parent)
  {
    RichText self = (RichText)this.self;
    this.model = self.model();
    if (model == null) throw Err.make("RichText.model is null").val;

    int style = self.multiLine.val ? SWT.MULTI: SWT.SINGLE;
    if (self.border.val)   style |= SWT.BORDER;
    if (self.wrap.val)     style |= SWT.WRAP;
    if (self.hscroll.val)  style |= SWT.H_SCROLL;
    if (self.vscroll.val)  style |= SWT.V_SCROLL;

    StyledText t = new StyledText((Composite)parent, style);
    control = t;
    t.setContent(content = new Content());
    t.addLineStyleListener(this);
    t.addModifyListener(this);
    return t;
  }

  public void sync(Str f)
  {
    Env env = Env.get();
    RichText self = (RichText)this.self;
    StyledText c = (StyledText )this.control;

    if (f == null || f == self.fontId)  c.setFont(env.font(self.font()));
  }

  public void modifyText(ModifyEvent event)
  {
    RichText self = (RichText )this.self;
    if (self.onModify != null) self.onModify.call1(event(event));
  }

//////////////////////////////////////////////////////////////////////////
// Content
//////////////////////////////////////////////////////////////////////////

  class Content implements StyledTextContent
  {
    // Return the number of characters in the content.
    public int getCharCount()
    {
      return (int)model.charCount().val;
    }

    // Return the number of lines.
    public int getLineCount()
    {
      return (int)model.lineCount().val;
    }

    // Return the line at the given line index without delimiters.
    public String getLine(int lineIndex)
    {
      return model.line(Int.pos(lineIndex)).val;
    }

    // Return the line index at the given character offset.
    public int getLineAtOffset(int offset)
    {
      return (int)model.lineAtOffset(Int.pos(offset)).val;
    }

    // Return the character offset of the first character of the given line.
    public int getOffsetAtLine(int lineIndex)
    {
      return (int)model.offsetAtLine(Int.pos(lineIndex)).val;
    }

    // Return the line delimiter that should be used by the StyledText widget when inserting new lines.
    public String getLineDelimiter()
    {
      return model.lineDelimiter().val;
    }

    // Returns a string representing the content at the given range.
    public String getTextRange(int start, int len)
    {
      return model.textRange(Int.pos(start), Int.pos(len)).val;
    }

    // Replace the text with "newText" starting at position "start" for a length of "len".
    public void replaceTextRange(int start, int len, String newText)
    {
      // TODO
      int newLineCount = 0;
      for (int i=0; i<newText.length(); ++i)
        if (newText.charAt(i) == '\n') newLineCount++;

      int oldLineCount = 0;
      String old = getTextRange(start, len);
      for (int i=0; i<old.length(); ++i)
        if (old.charAt(i) == '\n') oldLineCount++;

      TextChangingEvent event = new TextChangingEvent(this);
      event.start = start;
      event.replaceCharCount = len;
      event.replaceLineCount = oldLineCount;
      event.newText = newText;
      event.newCharCount = newText.length();
      event.newLineCount = newLineCount;

      //System.out.println("##### replace: " + start + " len=" + len + " " + Str.make(newText).toCode());
      //System.out.println("---> start=" + event.start + " replaceCharCount="+ event.replaceCharCount + " replaceLineCount=" + event.replaceLineCount
      //                   + " newCharCount="+ event.newCharCount+ " newLineCount=" + event.newLineCount);

      for (int i=0; i<listeners.size(); ++i)
        ((TextChangeListener)listeners.get(i)).textChanging(event);

      model.replaceTextRange(Int.pos(start), Int.pos(len), Str.make(newText));


      TextChangedEvent event2 = new TextChangedEvent (this);
      for (int i=0; i<listeners.size(); ++i)
        ((TextChangeListener)listeners.get(i)).textChanged(event2);
    }

    // Set text to "text".
    public void setText(String text)
    {
      model.text(Str.make(text));
    }

    // Called by StyledText to add itself as an Observer to content changes.
    public void addTextChangeListener(TextChangeListener listener)
    {
      listeners.add(listener);

      TextChangedEvent event = new TextChangedEvent(this);
      for (int i=0; i<listeners.size(); ++i)
        ((TextChangeListener)listeners.get(i)).textSet(event);
    }

    // Remove the specified text changed listener.
    public void removeTextChangeListener(TextChangeListener listener)
    {
      listeners.remove(listener);
    }

    ArrayList listeners = new ArrayList();
  }

//////////////////////////////////////////////////////////////////////////
// Styling
//////////////////////////////////////////////////////////////////////////

  public void lineGetStyle(LineStyleEvent event)
  {
    Env env = Env.get();
    int lineOffset = event.lineOffset;
    int lineLen = event.lineText.length();

    // get Int/RichTextStyle list where Int is offset in line
    List list = model.lineStyles(model.lineAtOffset(Int.pos(lineOffset)));
    if (list == null) return;

    // map Int/RichTextStyle list to StyleRange array
    int num = list.sz()/2;
    StyleRange[] results = event.styles = new StyleRange[num];
    for (int i=0; i<num; ++i)
    {
      Int offsetInLine = (Int)list.get(i*2);
      RichTextStyle s = (RichTextStyle)list.get(i*2+1);
      results[i] = toStyleRange(env, s, lineOffset+(int)offsetInLine.val);
      if (i > 0) results[i-1].length = results[i].start - results[i-1].start;
    }
    if (num > 0)
      results[num-1].length = lineLen - (results[num-1].start - lineOffset);
  }

  StyleRange toStyleRange(Env env, RichTextStyle s, int start)
  {
    StyleRange r = new StyleRange();
    r.start          = start;
    r.foreground     = env.color(s.fg);
    r.background     = env.color(s.bg);
    r.font           = env.font(s.font);

    /* waiting for Eclipse 3.4...
    if (s.underlineStyle != null && s.underlineStyle != Int.Zero)
    {
      switch (s.underlineStyle.val)
      {
        case 1: r.underlineStyle = SWT.UNDERLINE_SINGLE;   break;
        case 2: r.underlineStyle = SWT.UNDERLINE_DOUBLE;   break;
        case 3: r.underlineStyle = SWT.UNDERLINE_ERROR;    break;
        case 4: r.underlineStyle = SWT.UNDERLINE_SQUIGGLE; break;
      }
    }
    */

    return r;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  RichTextModel model;
  Content content;
}