//
// Copyright (c) 2008, John Sublett
// Licensed under the Academic Free License version 3.0
//
// History:
//   13 Jan 08  John Sublett  Creation
//

using sql

**
** TableDef defines the mapping of a fan type to a database table.
**
const internal class TableDef
{
  **
  ** Make a new instance that maps the specified type.
  **
  new make(HavenService haven, Type rowType)
  {
    if (!rowType.facet("serializable", false))
      throw HavenErr.make("Persistent types must be serializable.")

    this.haven = haven
    this.rowType = rowType.toImmutable
    this.tableName = haven.tableName(rowType)
    mColumns := ColDef[,]
    mColumnMap := Str:ColDef[:]
    mKeyColumns := ColDef[,]
    mNonKeyColumns := ColDef[,]
    mAutoColumns := ColDef[,]
    idefs := Str:IndexDef[][:]
    rowType.fields.each |Field f|
    {
      if (f.facet("transient", false)) return
      col := haven.makeColumn(f)
      mColumns.add(col)
      mColumnMap[f.name] = col
      if (col.isKey)
        mKeyColumns.add(col)
      else
        mNonKeyColumns.add(col)
      if (col.isAuto) mAutoColumns.add(col)
      
      index := col.index
      if (index != null)
      {
        i := idefs[index.name]
        if (i == null)
          idefs[index.name] = [index]
        else
          i.add(index)
      }
    }
    
    // collapse index definitions
    mIndices := IndexDef[,]
    idefs.values.each |IndexDef[] defs|
    {
      mIndices.add(IndexDef.merge(defs))
    }
    
    defaultRow := rowType.make
    mColumns.each |ColDef col|
    {
      col.validate(defaultRow)
    }
    
    columns = mColumns.toImmutable
    columnMap = mColumnMap.toImmutable
    keyColumns = mKeyColumns.toImmutable
    nonKeyColumns = mNonKeyColumns.toImmutable
    autoColumns = mAutoColumns.toImmutable
    indices = mIndices.toImmutable
  }

  **
  ** Get the DDL required to create this table in the associated
  ** database.
  **
  internal Str[] createTableDdl()
  {
    stmts := Str[,]
    
    ddl := StrBuf.make(256)
    ddl.add("CREATE TABLE ")
    ddl.add(tableName)
    keys := ColDef[,]
    refs := ColDef[,]
    
    // columns
    if (!columns.isEmpty)
    {
      ddl.add(" (")
      columns.each |ColDef c, Int index|
      {
        if (index != 0) ddl.add(", ")
        ddl.add(c.definition)
        
        if (c.isKey) keys.add(c)
        if (c.isRef) refs.add(c)
      }
    }
    
    // primary key
    if (!keys.isEmpty)
    {
      ddl.add(", PRIMARY KEY (")
      keys.each |ColDef c, Int index|
      {
        if (index != 0) ddl.add(", ")
        ddl.add(c.name)
      }
      ddl.add(")")
    }
    
    // foreign keys
    if (!refs.isEmpty)
    {
      refs.each |ColDef c, Int index|
      {
        ddl.add(", FOREIGN KEY ($c.name) REFERENCES ").add(c->foreignKey)
      }
    }

    ddl.add(");")
    stmts.add(ddl.toStr)
    
    if (indices.size != 0)
    {
      indices.each |IndexDef index|
      {
        ddl.clear
        ddl.add("CREATE INDEX ").add(index.name)
        ddl.add(" ON ").add(tableName)
        ddl.add(" (")
        index.columns.each |ColDef col, Int i|
        {
          if (i != 0) ddl.add(", ")
          ddl.add(col.columnName)
          if (index.desc) ddl.add(" DESC")
        }
        ddl.add(")")
        ddl.add(";")
        stmts.add(ddl.toStr)
      }
    }
    
    return stmts
  }
  
  **
  ** Get the DDL required to drop this table in the associated
  ** database.
  **
  internal Str deleteTableDdl()
  {
    ddl := StrBuf.make(128)
    ddl.add("DROP TABLE ")
    ddl.add(tableName)
    return ddl.toStr
  }

//////////////////////////////////////////////////////////////////////////
// Insert
//////////////////////////////////////////////////////////////////////////

  **
  ** Get the DML required to insert the specified object in the database.
  **
  internal Str insertDml(Obj row)
  {
    dml := StrBuf.make(128)
    dml.add("INSERT INTO ")
    dml.add(tableName)
    dml.add(" (")
    index := 0
    columns.each |ColDef col|
    {
      if (col.isAuto) return
      if (index != 0) dml.add(", ")
      dml.add(col.name)
      index++
    }
    dml.add(") VALUES (")
    index = 0
    columns.each |ColDef col|
    {
      if (col.isAuto) return
      if (index != 0) dml.add(", ")
      dml.add(col.fieldToSql(row))
      index++
    }
    dml.add(")")
    
    //echo("TableDef.insertDml: " + dml)
    
    return dml.toStr
  }
  
  **
  ** Get the DML required to insert the specified object in the database.
  **
  internal Statement prepareInsert()
  {
    dml := StrBuf.make(128)
    dml.add("INSERT INTO ")
    dml.add(tableName)
    dml.add(" (")
    index := 0
    columns.each |ColDef col|
    {
      if (col.isAuto) return
      if (index != 0) dml.add(", ")
      dml.add(col.columnName)
      index++
    }
    dml.add(") VALUES (")
    index = 0
    columns.each |ColDef col|
    {
      if (col.isAuto) return
      if (index != 0) dml.add(", ")
      dml.add("@$col.field.name")
      index++
    }
    dml.add(")")
    
    return haven.sql(dml.toStr).prepare
  }
  
  **
  ** Update the autogenerated field values for the
  ** specified 'newRow' object.  The newRow must
  ** be the last object created in the database.
  **
  internal Void updateAutoColumns(Obj newRow)
  {
    if (autoColumns.isEmpty) return

    if (autoColumns.size > 1)
      throw HavenErr.make("Only one autogenerated key is supported.")
    
    autoKeyCol := autoColumns[0]
    autoKeyCol.field.set(newRow, haven.lastAutoKey)
  }

  **
  ** Get the key value from the specified object.
  **
  internal Obj keyValue(Obj row)
  {
    if (row.type != rowType)
      throw HavenErr.make("Type mismatch: $row.type != $rowType")
    
    if (keyColumns.size != 1)
      throw HavenErr.make("Multi-field keys are not supported.")
    
    return keyColumns[0].field.get(row)
  }

//////////////////////////////////////////////////////////////////////////
// Read
//////////////////////////////////////////////////////////////////////////

  **
  ** Get the SQL for reading all objects from the table.
  **
  internal Str readAllSql()
  {
    echo("readAllSql")
    sql := StrBuf.make(128)
    sql.add("SELECT ")
    columns.each |ColDef col, Int index|
    {
      if (index != 0) sql.add(", ")
      sql.add(col.columnName)
    }
    sql.add(" FROM ").add(tableName)
    
    return sql.toStr
  }
  
  internal Statement prepareReadAllSql()
  {
    sql := StrBuf.make(128)
    sql.add("SELECT ")
    columns.each |ColDef col, Int index|
    {
      if (index != 0) sql.add(", ")
      sql.add(col.columnName)
    }
    sql.add(" FROM ").add(tableName)
    
    return haven.sql(sql.toStr).prepare
  }

  **
  ** Get the SQL for reading an object from this table using the
  ** specified id value.
  **
  internal Str readByIdSql(Obj id)
  {
    if (keyColumns.size != 1)
    {
      if (keyColumns.size > 1)
        throw HavenErr.make("$rowType.qname has multiple key fields.")
      if (keyColumns.size == 0)
        throw HavenErr.make("$rowType.qname does not have a key column.")
    }
    return readByIdMapSql([keyColumns[0].columnName:id])
  }
  
  **
  ** Prepare a statement for  reading an object from this table using the
  ** specified id value.
  **
  internal Statement prepareReadById()
  {
    if (keyColumns.size != 1)
    {
      if (keyColumns.size > 1)
        throw HavenErr.make("$rowType.qname has multiple key fields.")
      if (keyColumns.size == 0)
        throw HavenErr.make("$rowType.qname does not have a key column.")
    }
    
    sql := StrBuf.make(128)
    sql.add("SELECT ")
    columns.each |ColDef col, Int index|
    {
      if (index != 0) sql.add(", ")
      sql.add(col.columnName)
    }
    sql.add(" FROM ").add(tableName)
    sql.add(" WHERE ")
    keyColumns.each |ColDef keyCol, Int index|
    {
      if (index != 0) sql.add(" AND ")
      sql.add(keyCol.columnName).add(" = @$keyCol.field.name")
    }
    
    return haven.sql(sql.toStr).prepare
  }

  **
  ** Get the SQL for reading an object from this table using
  ** he specified named values.
  **
  internal Str readByIdMapSql(Str:Obj ids)
  {
    sql := StrBuf.make(128)
    sql.add("SELECT ")
    columns.each |ColDef col, Int index|
    {
      if (index != 0) sql.add(", ")
      sql.add(col.columnName)
    }
    sql.add(" FROM ").add(tableName)
    sql.add(" WHERE ")
    ids.keys.each |Str id, Int index|
    {
      if (index != 0) sql.add(" AND ")
      sql.add(id).add(" = ").add(columnMap[id].toSql(ids[id]))
    }
    
    return sql.toStr
  }

  **
  ** Create an object from the specified table row.
  **
  internal Obj fromRow(Row row, HavenNamespace ns)
  {
    obj := rowType.make
    columns.each |ColDef col|
    {
      col.setFieldFromRow(obj, row, ns)
    }
    return obj
  }

//////////////////////////////////////////////////////////////////////////
// Update
//////////////////////////////////////////////////////////////////////////

  **
  ** Get the DML required to update the specified database object.
  **
  internal Str updateDml(Obj obj)
  {
    dml := StrBuf.make(256)
    dml.add("UPDATE ").add(tableName)
    dml.add(" SET ")
    nonKeyColumns.each |ColDef col, Int index|
    {
      if (index != 0) dml.add(", ")
      dml.add(col.columnName).add(" = ").add(col.fieldToSql(obj))
    }
    dml.add(" WHERE ")
    keyColumns.each |ColDef col, Int index|
    {
      if (index != 0) dml.add(" AND ")
      dml.add(col.columnName).add(" = ").add(col.fieldToSql(obj))
    }
    
    return dml.toStr
  }
  
  **
  ** Prepare a statement to update the specified database object.
  **
  internal Statement prepareUpdate()
  {
    dml := StrBuf.make(256)
    dml.add("UPDATE ").add(tableName)
    dml.add(" SET ")
    nonKeyColumns.each |ColDef col, Int index|
    {
      if (index != 0) dml.add(", ")
      dml.add(col.columnName).add(" = @").add(col.field.name)
    }
    dml.add(" WHERE ")
    keyColumns.each |ColDef col, Int index|
    {
      if (index != 0) dml.add(" AND ")
      dml.add(col.columnName).add(" = @$col.field.name")
    }
    
    return haven.sql(dml.toStr).prepare
  }

//////////////////////////////////////////////////////////////////////////
// Delete
//////////////////////////////////////////////////////////////////////////

  **
  ** Get the DML required to delete the specified object from
  ** the database.
  **
  internal Str deleteDml(Obj obj)
  {
    dml := StrBuf.make(128)
    dml.add("DELETE FROM ").add(tableName)
    dml.add(" WHERE ")
    keyColumns.each |ColDef col, Int index|
    {
      if (index != 0) dml.add(" AND ")
      dml.add(col.columnName).add(" = ").add(col.fieldToSql(obj))
    }
    
    return dml.toStr
  }
  
  **
  ** Prepare a statement to delete the specified object from
  ** the database.
  **
  internal Statement prepareDelete()
  {
    dml := StrBuf.make(128)
    dml.add("DELETE FROM ").add(tableName)
    dml.add(" WHERE ")
    keyColumns.each |ColDef col, Int index|
    {
      if (index != 0) dml.add(" AND ")
      dml.add(col.columnName).add(" = @$col.field.name")
    }
    
    return haven.sql(dml.toStr).prepare
  }
  
  **
  ** Get the DML required to delete the specified object from
  ** the database.
  **
  internal Str deleteDmlById(Obj id)
  {
    dml := StrBuf.make(128)
    dml.add("DELETE FROM ").add(tableName)
    dml.add(" WHERE ")
    idCol := keyColumns[0]
    dml.add(idCol.columnName).add(" = ").add(idCol.toSql(id))
    
    return dml.toStr
  }
  
  **
  ** Prepare a statement to delete the specified object from
  ** the database.
  **
  internal Statement prepareDeleteById()
  {
    dml := StrBuf.make(128)
    dml.add("DELETE FROM ").add(tableName)
    dml.add(" WHERE ")
    idCol := keyColumns[0]
    dml.add(idCol.columnName).add(" = @$idCol.field.name")
    
    return haven.sql(dml.toStr).prepare
  }

  **
  ** Get the DML required to delete all objects from this table.
  **
  internal Str deleteAllDml()
  {
    dml := StrBuf.make(128)
    dml.add("DELETE FROM ").add(tableName)
    return dml.toStr
  }
  
  **
  ** Prepare a statement to delete all objects from this table.
  **
  internal Statement prepareDeleteAll()
  {
    dml := StrBuf.make(128)
    dml.add("DELETE FROM ").add(tableName)
    return haven.sql(dml.toStr).prepare
  }

//////////////////////////////////////////////////////////////////////////
// Prepared Statements
//////////////////////////////////////////////////////////////////////////

  internal Int execute(Statement stmt, Obj obj)
  {
    params := Str:Obj[:]
    columns.each |ColDef col|
    {
      col.paramValue(col.field.get(obj), params)
    }
    
    return stmt.execute(params)
  }
  
  internal Str:Obj idParam(Obj id)
  {
    params := Str:Obj[:]
    keyColumns[0].paramValue(id, params)
    return params
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  const HavenService haven

  **
  ** The type for the rows in the table.
  **
  const Type rowType

  **
  ** The name of the table in the database.
  **
  const Str tableName

  **
  ** The map of columns by name.
  **
  const Str:ColDef columnMap

  **
  ** The list of columns in the table.
  **
  const ColDef[] columns

  **
  ** The list of columns whose values compose
  ** the primary key.
  **
  const ColDef[] keyColumns
  
  **
  ** The list of columns whose values are not
  ** part of the primary key.
  **
  const ColDef[] nonKeyColumns
  
  **
  ** The list of columns whose values are autogenerated
  ** by the database.
  **
  const ColDef[] autoColumns

  **
  ** The list of indices on this table.
  **
  const IndexDef[] indices

}