//
// Copyright (c) 2008, John Sublett
// Licensed under the Academic Free License version 3.0
//
// History:
//   05 Feb 08  John Sublett  Creation
//

using sql

**
** Test the object-relational mapping features of HavenService.
**
class HavenTest : Test
{

  once HavenService haven()
  {
    return HavenService.make(
      null,
      Sys.env["sql.test.connection"],
      Sys.env["sql.test.username"],
      Sys.env["sql.test.password"],
      Type.find(Sys.env["sql.test.dialect"]).make)
  }

  Void test()
  {
    haven.start

    try
    {
      if (!verifyOpen) return

      haven.open
      dropTables(haven)
      haven.close

      createTableErrors
      createTable
      create
      bufTest(BigBuf#, 30000)
      clobTest(BigClob#, 5)
      autoIncrement
      read
      list
      update
      delete
      ref
      index
      namespace
    }
    finally
    {
      haven.stop
    }
  }

//////////////////////////////////////////////////////////////////////////
// Open
//////////////////////////////////////////////////////////////////////////

  Bool verifyOpen()
  {
    try
    {
      haven.open
      haven.close
      return true
    }
    catch (Err e)
    {
      echo("**")
      echo("** WARNING: Cannot perform HavenTest without available database: " + haven.type.qname)
      echo("**          $e")
      echo("**")
      return false
    }
  }

  **
  ** Test expected table creation errors.
  **
  Void createTableErrors()
  {
    haven.open

    verifyErr(HavenErr#) |,|
    {
      haven.createTable(StrNoMax#)
    }

    //verifyErr(HavenErr#) |,|
    //{
    //  haven.createTable(NotNullNoDefault#)
    //}

    verifyErr(HavenErr#) |,|
    {
      haven.createTable(AutoStr#)
    }

    verifyErr(HavenErr#) |,|
    {
      haven.createTable(NotSerializable#)
    }

    haven.close
  }

  **
  ** Test successful table creation.
  **
  Void createTable()
  {
    haven.open

    haven.createTable(TestRecord#)
    tableName := haven.tableName(TestRecord#)
    verify(haven.tableExists(tableName), "$tableName was not created")
    verify(haven.typeTableExists(TestRecord#))
    row := haven.tableRow(tableName)
    rowFields := row.type.fields
    typeFields := TestRecord#.fields
    verifyEq(rowFields.size, typeFields.size)
    for (i:=0; i<typeFields.size; i++)
      verifyEq(rowFields[i].name, typeFields[i].name)

    verify(!haven.typeTableExists(Fish#))

    haven.close
  }

  **
  ** Test database object creation.
  **
  Void create()
  {
    haven.open

    haven.create(TestRecord.make)
    verifyEq(haven.sql("select * from " + haven.tableName(TestRecord#)).query.size, 1)
    row := haven.sql("select * from " + haven.tableName(TestRecord#)).query[0]
    verifyEq(row->s, "")
    verifyEq(row->i, 57)
    verifyEq(row->f, 3.14f)
    verifyEq(row->b, "t")
    verifyEq(row->d, 0)
    verifyEq(row->k, MeddlingKid.shaggy.ordinal)

    haven.sql("delete from " + haven.tableName(TestRecord#)).execute
    verifyEq(haven.sql("select * from " + haven.tableName(TestRecord#)).query.size, 0)
    newObj := TestRecord.make
    newObj.s = "haven"
    newObj.i = 66
    newObj.f = 2.71f
    newObj.b = false
    newObj.d = DateTime.makeTicks(1)
    newObj.k = MeddlingKid.daphne
    haven.create(newObj)
    verifyEq(haven.sql("select * from " + haven.tableName(TestRecord#)).query.size, 1)
    row = haven.sql("select * from " + haven.tableName(TestRecord#)).query[0]
    verifyEq(row->s, "haven")
    verifyEq(row->i, 66)
    verifyEq(row->f, 2.71f)
    verifyEq(row->b, "f")
    verifyEq(row->d, 1)
    verifyEq(row->k, MeddlingKid.daphne.ordinal)

    newObj = TestRecord.make { i = 100 }
    newObj.s = "quote'test"
    haven.create(newObj)
    readObj := (TestRecord)haven.read(TestRecord#, newObj.i)
    verifyEq(readObj->s, newObj.s)

    newObj = TestRecord.make { i = 101 }
    newObj.s = "a\nb\nc"
    haven.create(newObj)
    readObj = (TestRecord)haven.read(TestRecord#, newObj.i)
    verifyEq(readObj->s, newObj.s)

    newObj = TestRecord.make { i = 102 }
    newObj.s = "\""
    haven.create(newObj)
    readObj = (TestRecord)haven.read(TestRecord#, newObj.i)
    verifyEq(readObj->s, newObj.s)

    haven.close
  }

  Void bufTest(Type objType, Int size)
  {
    haven.open

    haven.createTable(objType)

    Buf testBuf := makeBuf(size)
    newObj := objType.make
    newObj->buf = testBuf
    haven.create(newObj)
    readObj := haven.read(objType, newObj->i)
    allMatch := true
    size.times |Int j|
    {
      if (((Buf)readObj->buf)[j] != testBuf[j]) allMatch = false
    }
    verify(allMatch)

    haven.close
  }

  Void clobTest(Type objType, Int size)
  {
    haven.open

    haven.createTable(objType)

    Str testClob := makeClob(size)
    newObj := objType.make
    newObj->clob = testClob
    haven.create(newObj)
    readObj := haven.read(objType, newObj->i)
    allMatch := true
    size.times |Int j|
    {
      if (((Str)readObj->clob)[j] != testClob[j]) allMatch = false
    }
    verify(allMatch)

    haven.close
  }

  **
  ** Test creation of database object with auto-increment field.
  **
  Void autoIncrement()
  {
    haven.open

    haven.createTable(AutoInc#)
    for (i:=0; i<10; i++)
    {
      obj := AutoInc.make
      haven.create(obj)
      verifyEq(obj.x, i+1)
    }

    verifyEq(haven.sql("select * from " + haven.tableName(AutoInc#)).query.size, 10)
    rows := haven.sql("select x from " + haven.tableName(AutoInc#) + " order by x").query
    rows.each |Obj row, Int index| { verifyEq(row->x, index+1) }

    haven.close
  }

  **
  ** Test read for database objects.
  **
  Void read()
  {
    haven.open

    row := TestRecord.make
    row.d = DateTime.now
    haven.create(row)
    rrow := haven.read(TestRecord#, 1234)
    verify(rrow == null)
    rrow = haven.read(TestRecord#, row.i)
    verifyEq(TestRecord#, rrow.type)
    TestRecord#.fields.each |Field f|
    {
      verifyEq(f.get(row), f.get(rrow))
    }

    haven.close
  }

  **
  ** Test list for database objects.
  **
  Void list()
  {
    haven.open

    haven.deleteAll(TestRecord#)

    records := TestRecord[,]
    for (i:=0; i<10; i++)
    {
      rec := TestRecord.make
      rec.i = i
      records.add(rec)
      haven.create(rec)
    }

    reclist := haven.listObjs(TestRecord#)
    verifyEq(reclist.size, records.size)

    haven.close
  }

  **
  ** Test update for database objects.
  **
  Void update()
  {
    haven.open

    haven.deleteAll(TestRecord#)
    rec1 := TestRecord.make
    {
      i = 1
      f = 1.21f
      b = true
      s = "filibuster"
      d = DateTime.now
      k = MeddlingKid.velma
    }
    haven.create(rec1)
    rec2 := haven.read(TestRecord#, 1)
    TestRecord#.fields.each |Field f|
    {
      verifyEq(f.get(rec1), f.get(rec2))
    }

    rec2->s = "pocket veto"
    rec2->d = DateTime.now
    haven.update(rec2)
    rec3 := haven.read(TestRecord#, 1)
    TestRecord#.fields.each |Field f|
    {
      verifyEq(f.get(rec2), f.get(rec3))
    }

    rec1 = rec2
    rec1->s = "update'q't"
    haven.update(rec1)
    rec2 = haven.read(TestRecord#, 1)
    TestRecord#.fields.each |Field f|
    {
      verifyEq(f.get(rec1), f.get(rec2))
    }

    haven.createTable(Test2#)
    t1 := Test2.make { s = "abcdefg" }
    haven.create(t1)
    t2 := haven.read(Test2#, t1.id) as Test2
    verifyEq(t1.s, t2.s)

    t1.s = "hijklmnop"
    haven.update(t1)
    t2 = haven.read(Test2#, t2.id) as Test2
    verifyEq(t1.s, t2.s)

    haven.close
  }

  **
  ** Test delete for database objects.
  **
  Void delete()
  {
    haven.open

    haven.deleteAll(TestRecord#)
    verify(haven.read(TestRecord#, 1) == null)
    verify(haven.read(TestRecord#, 2) == null)

    row1 := TestRecord.make
    row1.i = 1
    haven.create(row1)
    row2 := TestRecord.make
    row2.i = 2
    haven.create(row2)
    verify(haven.read(TestRecord#, 1) != null)
    verify(haven.read(TestRecord#, 2) != null)

    verify(haven.delete(row1))
    verify(haven.read(TestRecord#, 1) == null)
    verify(haven.read(TestRecord#, 2) != null)
    verify(!haven.delete(row1))

    verify(haven.delete(row2))
    verify(haven.read(TestRecord#, 1) == null)
    verify(haven.read(TestRecord#, 2) == null)

    // delete by id
    row1 = TestRecord.make
    row1.i = 66
    haven.create(row1)
    verify(haven.read(TestRecord#, 66) != null)
    haven.deleteById(TestRecord#, 66)
    verify(haven.read(TestRecord#, 66) == null)

    haven.close
  }

  **
  ** Test persistent types with uri refs.
  **
  Void ref()
  {
    haven.open

    verifyErr(SqlErr#) |,| { haven.createTable(Fish#) }
    haven.createTable(Pond#)
    haven.createTable(Fish#)

    pond := Pond.make("Froggy Bottom")
    pondUri := haven.create(pond)
    verifyEq(pond.id, 1)

    nemo := Fish.make("nemo", 0.5f)
    nemo.pond = pondUri
    haven.create(nemo)
    Obj[] fishPonds :=
      haven.sql("select pond from " + haven.tableName(Fish#) + " where name = 'nemo'").query
    verifyEq(fishPonds.size, 1)
    verifyEq(fishPonds[0]->pond, 1)

    fish := haven.read(Fish#, "nemo") as Fish
    verifyEq(fish.name, "nemo")
    verifyEq(fish.pond, nemo.pond)

    haven.close
  }

  **
  ** Test column indexing.
  **
  Void index()
  {
    haven.open

    tdef := haven.table(Indexed#)
    verifyEq(tdef.indices.size, 4)
    tdef.indices.each |IndexDef idef|
    {
      if (idef.name == "haven_Indexed_a")
      {
        verifyEq(idef.columns.size, 1)
        verifyEq(idef.columns[0].columnName, "a")
        verify(!idef.desc)
      }
      else if (idef.name == "haven_Indexed_bc")
      {
        verifyEq(idef.columns.size, 2)
        verifyEq(idef.columns[0].columnName, "b")
        verifyEq(idef.columns[1].columnName, "c")
        verify(!idef.desc)
      }
      else if (idef.name == "haven_Indexed_d")
      {
        verifyEq(idef.columns.size, 1)
        verifyEq(idef.columns[0].columnName, "d")
        verify(idef.desc)
      }
      else if (idef.name == "haven_Indexed_ef")
      {
        verifyEq(idef.columns.size, 2)
        verifyEq(idef.columns[0].columnName, "e")
        verifyEq(idef.columns[1].columnName, "f")
        verify(idef.desc)
      }
    }

    Err ex
    try
    {
      haven.createTable(Indexed#)
    }
    catch(Err exx)
    {
      ex = exx
      exx.trace
    }
    verify(ex == null, "createTable failed")

    haven.close
  }

  Void namespace()
  {
    haven.open

    ns := haven.ns
    Sys.mount(`/haven`, ns)

    t1 := TestRecord.make
    t1.f = 1.21f
    verifyErr(ArgErr#) |,| { ns.create(`/a/b/c`, t1) }
    t1Uri := ns.create(null, t1)
    verifyEq(`/haven/haven/TestRecord/57`, t1Uri)

    t2 := (TestRecord)`/haven/haven/TestRecord/57`.get
    verify(t2.f == 1.21f)

    t1.f = 8675309f
    ns.put(`/haven/haven/TestRecord/57`, t1)
    t2 = (TestRecord)`/haven/haven/TestRecord/57`.get
    verify(t2.f == 8675309f)
    t1.i = 99
    verifyErr(HavenErr#) |,| { ns.put(`/haven/haven/TestRecord/57`, t1) }

    verifyErr(UnresolvedErr#) |,| { ns.delete(`/haven/haven/TestRecord/1234`) }
    ns.delete(`/haven/haven/TestRecord/57`)
    verifyEq(ns.get(`/haven/haven/TestRecord/57`, false), null)

    pond := Pond.make("Lilly Pad Lake")
    pondUri := ns.create(null, pond)
    verifyEq(pondUri, ("/haven/haven/Pond/" + pond.id).toUri)

    fish := Fish.make("Swimmy")
    fish.pond = pondUri
    fishUri := ns.create(null, fish)

    fish2 := (Fish)fishUri.get
    verifyEq(fish2.pond, pondUri)
    pond2 := fish2.pond.get as Pond

    Pond#.fields.each |Field f|
    {
      verifyEq(f.get(pond), f.get(pond2))
    }

    ObjTable pondTable := `/haven/haven/TestRecord`.get
    verifyEq(pondTable.objType, TestRecord#)

    haven.close
  }

//////////////////////////////////////////////////////////////////////////
// Util
//////////////////////////////////////////////////////////////////////////

  Buf makeBuf(Int size)
  {
    b := Buf.make(size)
    size.times |Int i| { b.write(i) }
    return b
  }

  Str makeClob(Int size)
  {
    StrBuf s := StrBuf.make(size)
    size.times |Int i| { s.add("a") }
    return s.toStr
  }

  **
  ** Drop all tables in the database
  **
  Void dropTables(SqlService db)
  {
    Str[] tables := haven.tables.dup
    while (tables.size != 0)
    {
      Int dropped := 0
      tables.each |Str tableName|
      {
        try
        {
          haven.sql("drop table $tableName").execute
          tables.remove(tableName)
          dropped++
        }
        catch(Err e)
        {
        }
      }

      if (dropped == 0)
        throw HavenErr.make("All tables could not be dropped.")
    }
  }

}
